use std::fmt::Display;

use crate::{storage, Amount, Asset, Status, TransactionId};
use serde::{Serialize, Serializer};

#[derive(thiserror::Error, Debug, Serialize)]
pub enum Error {
    #[error("Payment {0} is already spent")]
    SpentPayment(usize),

    #[error("Payment {0} is in the incorrect status {1}")]
    InvalidPaymentStatus(usize, Status),

    #[error("Spending {0:?} and issuing {1:?}. Both amounts must match")]
    InvalidAmount(Amount, Amount),

    #[error("Missing input payment for asset {0}")]
    MissingSpendingAsset(Asset),

    #[error("Missing output payment for asset {0}")]
    MissingPaymentAsset(Asset),

    #[error("Transaction updates are not allowed")]
    TransactionUpdatesNotAllowed,

    #[error("Mismatch between spend and create payment assets")]
    MismatchSpendingPayingAssets,

    #[error("Status transition from {0} to {1} is not allowed")]
    StatusTransitionNotAllowed(Status, Status),

    #[error("Storage: {0}")]
    Storage(#[from] storage::Error),

    #[error("Internal error at serializing: {0}")]
    #[serde(serialize_with = "serialize_to_string")]
    Internal(#[from] Box<bincode::ErrorKind>),

    #[error("Invalid calculated id {0} (expected {1})")]
    InvalidTransactionId(TransactionId, TransactionId),

    #[error("Overflow")]
    Overflow,
}

fn serialize_to_string<S, T>(value: &T, serializer: S) -> Result<S::Ok, S::Error>
where
    S: Serializer,
    T: ToString + Display,
{
    serializer.serialize_str(&value.to_string())
}
